/*
Copyright (C) 2006 - 2009 Evan Teran
                   eteran@alum.rit.edu

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
*/

#include "Analyzer.h"
#include "Debugger.h"
#include "State.h"
#include "Util.h"
#include "DebuggerCoreInterface.h"
#include "ArchProcessorInterface.h"
#include "SymbolManager.h"
#include "MemoryRegions.h"
#include "Instruction.h"
#include "AnalyzerOptionsPage.h"

#include <QtDebug>
#include <QMenu>
#include <QStack>
#include <QHash>
#include <QProgressDialog>
#include <QCoreApplication>
#include <QtDebug>
#include <QSettings>
#include <QMessageBox>

#include <boost/bind.hpp>

#ifdef USE_QT_CONCURRENT
#include <QtConcurrentMap>
#endif

#include <cstring>

#ifdef Q_OS_LINUX
#include <elf.h>
#endif

#define MIN_REFCOUNT 2

namespace {
#if defined(EDB_X86)
	const edb::Operand::Register STACK_REG = edb::Operand::REG_ESP;
	const edb::Operand::Register FRAME_REG = edb::Operand::REG_EBP;
#elif defined(EDB_X86_64)
	const edb::Operand::Register STACK_REG = edb::Operand::REG_RSP;
	const edb::Operand::Register FRAME_REG = edb::Operand::REG_RBP;
#endif
}

//------------------------------------------------------------------------------
// Name: Analyzer(
// Desc: 
//------------------------------------------------------------------------------
Analyzer::Analyzer() : menu_(0) {
}

//------------------------------------------------------------------------------
// Name: 
// Desc: 
//------------------------------------------------------------------------------
QWidget *Analyzer::options_page() {
	return new AnalyzerOptionsPage;
}

//------------------------------------------------------------------------------
// Name: menu(QWidget *parent)
// Desc: 
//------------------------------------------------------------------------------
QMenu *Analyzer::menu(QWidget *parent) {
	if(menu_ == 0) {
		menu_ = new QMenu(tr("Analyzer"), parent);
#if defined(EDB_X86)
		menu_->addAction(tr("&Analyzer EIP's Region"), this, SLOT(do_ip_analysis()), QKeySequence(tr("Ctrl+A")));
#elif defined(EDB_X86_64)
		menu_->addAction(tr("&Analyzer RIP's Region"), this, SLOT(do_ip_analysis()), QKeySequence(tr("Ctrl+A")));
#endif
		menu_->addAction(tr("&Analyzer Viewed Region"), this, SLOT(do_view_analysis()), QKeySequence(tr("Ctrl+Shift+A")));
	}

	return menu_;
}

//------------------------------------------------------------------------------
// Name: private_init()
// Desc: 
//------------------------------------------------------------------------------
void Analyzer::private_init() {
	edb::v1::set_analyzer(this);
}

//------------------------------------------------------------------------------
// Name: do_ip_analysis()
// Desc: 
//------------------------------------------------------------------------------
void Analyzer::do_ip_analysis() {
	MemRegion region;
	const State &state = edb::v1::current_state();
	const edb::address_t eip = state.instruction_pointer();
	if(edb::v1::memory_regions().find_region(eip, region)) {
		do_analysis(region);
	}
}

//------------------------------------------------------------------------------
// Name: do_view_analysis()
// Desc: 
//------------------------------------------------------------------------------
void Analyzer::do_view_analysis() {
	do_analysis(edb::v1::current_cpu_view_region());
}

//------------------------------------------------------------------------------
// Name: 
// Desc: 
//------------------------------------------------------------------------------
QList<QAction *> Analyzer::cpu_context_menu() {
	
	QList<QAction *> ret;
	
	QAction *const actFind = new QAction(tr("Analyze Here"), this);
	connect(actFind, SIGNAL(triggered()), this, SLOT(do_view_analysis()));
	ret << actFind;

	return ret;
}

//------------------------------------------------------------------------------
// Name: 
// Desc: 
//------------------------------------------------------------------------------
void Analyzer::do_analysis(const MemRegion &region) {
	if(region.size() != 0) {
		QProgressDialog progress(tr("Performing Analysis"), 0, 0, 100, edb::v1::debugger_ui);
		connect(this, SIGNAL(update_progress(int)), &progress, SLOT(setValue(int)));
		progress.show();
		progress.setValue(0);
		analyze(region);
		edb::v1::repaint_cpu_view();
	}
}

//------------------------------------------------------------------------------
// Name: 
// Desc: 
//------------------------------------------------------------------------------
void Analyzer::find_function_calls(const MemRegion &region, FunctionMap &found_functions) {
	static const edb::address_t page_size = edb::v1::debugger_core->page_size();

	const edb::address_t size_in_pages = region.size() / page_size;
	
	try {
		QVector<quint8> pages(size_in_pages * page_size);

		if(edb::v1::debugger_core->read_pages(region.start, &pages[0], size_in_pages)) {
			for(edb::address_t i = 0; i < size_in_pages * page_size; ++i) {

				const edb::Instruction insn(&pages[i], region.size() - i, region.start + i, std::nothrow);

				if(insn.valid() && insn.type() == edb::Instruction::OP_CALL) {

					const edb::address_t ip = region.start + i;
					const edb::Operand &op = insn.operand(0);

					if(op.general_type() == edb::Operand::TYPE_REL) {
						const edb::address_t ea = op.relative_target();

						// skip over ones which are : call <label>; label:
						if(ea != ip + insn.size()) {
							if(region.contains(ea)) {							
								// avoid calls which land in the middle of a function...
								// this may or may not be the best approach
								if(!is_inside_known(region, ea)) {
									found_functions[ea].entry_address	= ea;
									found_functions[ea].end_address		= ea;
									found_functions[ea].reference_count++;
								}
							}
						}
					}

					emit update_progress(util::percentage(5, 10, i, region.size()));
				}
			}
		}
	} catch(const std::bad_alloc &) {
		QMessageBox::information(0, tr("Memroy Allocation Error"),
			tr("Unable to satisfy memory allocation request for requested region."));
	}
}

//------------------------------------------------------------------------------
// Name: is_stack_frame(edb::address_t addr)
// Desc: 
//------------------------------------------------------------------------------
bool Analyzer::is_stack_frame(edb::address_t addr) const {

	quint8 buf[edb::Instruction::MAX_SIZE];

	unsigned int i = 0;

	while(i < 2) {
		// gets the bytes for the instruction
		int buf_size = sizeof(buf);
		if(!edb::v1::get_instruction_bytes(addr, buf, buf_size)) {
			break;
		}

		// decode it
		const edb::Instruction insn(buf, buf_size, addr, std::nothrow);
		if(!insn.valid()) {
			break;;
		}

		// which part of the sequence are we looking at?
		switch(i++) {
		case 0:
			// are we looking at a 'push rBP' ?
			if(insn.type() == edb::Instruction::OP_PUSH) {
				Q_ASSERT(insn.operand_count() == 1);
				const edb::Operand &op = insn.operand(0);
				if(op.complete_type() == edb::Operand::TYPE_REGISTER) {
					if(op.reg == FRAME_REG) {
						break;
					}
				}
			// if it is an 'enter 0,0' instruction, then it's a stack frame right away
			} else if(insn.type() == edb::Instruction::OP_ENTER) {
				Q_ASSERT(insn.operand_count() == 2);
				const edb::Operand &op0 = insn.operand(0);
				const edb::Operand &op1 = insn.operand(1);
				if(op0.immediate() == 0 && op1.immediate() == 0) {
					return true;
				}
			}
			++i;
			break;
		case 1:
			// are we looking at a 'mov rBP, rSP' ?
			if(insn.type() == edb::Instruction::OP_MOV) {
				Q_ASSERT(insn.operand_count() == 2);
				const edb::Operand &op0 = insn.operand(0);
				const edb::Operand &op1 = insn.operand(1);
				if(op0.complete_type() == edb::Operand::TYPE_REGISTER && op1.complete_type() == edb::Operand::TYPE_REGISTER) {
					if(op0.reg == FRAME_REG && op1.reg == STACK_REG) {
						return true;
					}
				}
			}
			break;
		default:
			break;
		}

		addr += insn.size();
	}
	
	return false;
}

//------------------------------------------------------------------------------
// Name: 
// Desc: 
//------------------------------------------------------------------------------
void Analyzer::bonus_stack_frames_helper(Function &info) const {
	const edb::address_t addr = info.entry_address;
	if(is_stack_frame(addr)) {
		info.reference_count++;
	}
}

//------------------------------------------------------------------------------
// Name: bonus_stack_frames(FunctionMap &results)
// Desc: give bonus if we see a "push ebp; mov ebp, esp;"
//------------------------------------------------------------------------------
void Analyzer::bonus_stack_frames(FunctionMap &results) {
#ifdef USE_QT_CONCURRENT
	QtConcurrent::blockingMap(results, boost::bind(&Analyzer::bonus_stack_frames_helper, this, _1));
#else
	std::for_each(results.begin(), results.end(), boost::bind(&Analyzer::bonus_stack_frames_helper, this, _1));
#endif
}

//------------------------------------------------------------------------------
// Name: 
// Desc: 
//------------------------------------------------------------------------------
void Analyzer::update_results_entry(FunctionMap &results, edb::address_t address) const {
	results[address].entry_address		= address;
	results[address].end_address		= address;
	if(results[address].reference_count == 0) {
		results[address].reference_count = MIN_REFCOUNT;
	} else {
		results[address].reference_count++;
	}
}

//------------------------------------------------------------------------------
// Name: bonus_main(const MemRegion &region, FunctionMap &results)
// Desc: 
//------------------------------------------------------------------------------
void Analyzer::bonus_main(const MemRegion &region, FunctionMap &results) const {
	const QString s = edb::v1::get_process_exe();
	if(!s.isEmpty()) {
		const edb::address_t main = edb::v1::locate_main_function();
		
		if(main != 0 && region.contains(main)) {
			// make sure we have an entry for this function
			update_results_entry(results, main);
		}
	}
}

//------------------------------------------------------------------------------
// Name: bonus_symbols(FunctionMap &results)
// Desc: 
//------------------------------------------------------------------------------
void Analyzer::bonus_symbols(const MemRegion &region, FunctionMap &results) {
	// give bonus if we have a symbol for the address
	
	QList<SymbolManager::SymbolP> symbols = edb::v1::symbol_manager().symbols();
	
	Q_FOREACH(const SymbolManager::SymbolP &sym, symbols) {
		const edb::address_t addr = sym->address;
		
		if(region.contains(addr)) {
			if(sym->is_code()) {

				qDebug("[Analyzer] adding: %s <%p>", qPrintable(sym->name), reinterpret_cast<void *>(addr));

				// make sure we have an entry for this function
				update_results_entry(results, addr);
			}
		}
	}
}

//------------------------------------------------------------------------------
// Name: 
// Desc: 
//------------------------------------------------------------------------------
int Analyzer::walk_all_functions(FunctionMap &results, const MemRegion &region, QSet<edb::address_t> &walked_functions) {
	int updates = 0;
	
	QSet<edb::address_t> found_functions;
		
	FunctionMap::iterator it = results.begin();
	while(it != results.end()) {
		Function &info = it.value();
		
		const FunctionMap::iterator next = ++it;

		if(info.reference_count >= MIN_REFCOUNT) {
			if(!walked_functions.contains(info.entry_address)) {
		
				// the function's upper bound is either the entry point of the next function
				// or the region's end which is the absolute max end this function can have
				const edb::address_t next_entry = (next != results.end()) ? next.value().entry_address : region.end;

				// walk the function and collect some results
				
				find_function_end(info, next_entry, found_functions, results);
				walked_functions.insert(info.entry_address);
				
	#if 1
				// if the very last instruction happens to be a jmp, then this may
				// be a call/ret -> jmp optimization. This isn't always the case
				// but often enough that it's probably right
				quint8 buf[edb::Instruction::MAX_SIZE];
				int buf_size = sizeof(buf);
				if(edb::v1::get_instruction_bytes(info.last_instruction, buf, buf_size)) {
					const edb::Instruction insn(buf, buf_size, info.last_instruction, std::nothrow);
					if(insn.valid() && insn.type() == edb::Instruction::OP_JMP) {

						Q_ASSERT(insn.operand_count() == 1);
						const edb::Operand &op = insn.operand(0);

						if(op.general_type() == edb::Operand::TYPE_REL) {
							const edb::address_t target = op.relative_target();

							Function func;

							if(!find_containing_function(target, func)) {
								found_functions.insert(target);
							}
						}
					}
				}
	#endif
			}
		}
		
		it = next;
	}
	
	// add the newly found functions to the list and report the number of "updates"
	Q_FOREACH(edb::address_t func, found_functions) {
		if(!results.contains(func)) {
			results[func].entry_address		= func;
			results[func].end_address		= func;
			results[func].reference_count	= MIN_REFCOUNT;
			++updates;
		}
	}
	
	return updates;
}

//------------------------------------------------------------------------------
// Name: fix_overlaps(FunctionMap &function_map)
// Desc: ensures that no function overlaps another
//------------------------------------------------------------------------------
void Analyzer::fix_overlaps(FunctionMap &function_map) {
	for(FunctionMap::iterator it = function_map.begin(); it != function_map.end(); ) {
		Function &func = *it++;
		if(it != function_map.end()) {
			const Function &next_func = *it;
			if(next_func.entry_address <= func.end_address) {				
				func.end_address = next_func.entry_address - 1;
			}
		}
	}
}

//------------------------------------------------------------------------------
// Name: 
// Desc: 
//------------------------------------------------------------------------------
void Analyzer::find_function_end(Function &function, edb::address_t end_address, QSet<edb::address_t> &found_functions, const FunctionMap &results) {
	QStack<edb::address_t>		jump_targets;
	QHash<edb::address_t, int>	visited_addresses;
		
	// we start with the entry point of the function
	jump_targets.push(function.entry_address);
	
	// while no more jump targets... (includes entry point)
	while(!jump_targets.empty()) {
		edb::address_t addr = jump_targets.pop();
				
		// for certain forward jump scenarioes this is possible.
		if(visited_addresses.contains(addr)) {
			continue;
		}
						
		// keep going until we go out of bounds	
		while(addr >= function.entry_address && addr < end_address) {
								
			quint8 buf[edb::Instruction::MAX_SIZE];
			int buf_size = sizeof(buf);
			if(!edb::v1::get_instruction_bytes(addr, buf, buf_size)) {
				break;
			}
			
			const edb::Instruction insn(buf, buf_size, addr, std::nothrow);
			if(!insn.valid()) {
				break;
			}

			// ok, it was a valid instruction, let's add it to the
			// list of possible 'code addresses'
			visited_addresses.insert(addr, insn.size());
			
			const edb::Instruction::Type type = insn.type();

			if(type == edb::Instruction::OP_RET || type == edb::Instruction::OP_HLT) {
				break;
			} else if(type == edb::Instruction::OP_JCC) {

				const edb::Operand &op = insn.operand(0);
				if(op.general_type() == edb::Operand::TYPE_REL) {
					const edb::address_t ea = op.relative_target();

					if(!visited_addresses.contains(ea) && !jump_targets.contains(ea)) {
						jump_targets.push(ea);
					}
				}
			} else if(type == edb::Instruction::OP_JMP) {

				const edb::Operand &op = insn.operand(0);
				if(op.general_type() == edb::Operand::TYPE_REL) {
					const edb::address_t ea = op.relative_target();
					
					// is it a jump to another function's entry point?
					// if so, this is a dead end, resume from other branches
					// but give the target a bonus reference
					FunctionMap::const_iterator it = results.find(ea);
					if(it != results.end()) {
						found_functions.insert(ea);
						break;
					}
					
					// ok, it is a jmp to an address we haven't seen yet
					// so just continue processing from there
					if(!visited_addresses.contains(ea)) {
						addr = ea;
						continue;
					} else {
						break;
					}
				}
			} else if(type == edb::Instruction::OP_CALL) {

				const edb::Operand &op = insn.operand(0);
				if(op.general_type() == edb::Operand::TYPE_REL) {
					const edb::address_t ea = op.relative_target();

					// skip over ones which are: "call <label>; label:"
					if(ea != addr + insn.size()) {
						found_functions.insert(ea);
					}
				} else if(op.general_type() == edb::Operand::TYPE_EXPRESSION) {
					// looks like: "call [...]", if it is of the form, call [N + REG]
					// then it may be a jump table using REG as an offset
					// TODO: is this a jump table?
				}
			}
			
			addr += insn.size();
		}
	}
	
	function.last_instruction	= function.entry_address;
	function.end_address		= function.entry_address;
	
	// get the last instruction and the last byte of the function
	for(QHash<edb::address_t, int>::const_iterator it = visited_addresses.begin(); it != visited_addresses.end(); ++it) {
		function.end_address        = qMax(function.end_address,      it.key() + it.value() - 1);
		function.last_instruction	= qMax(function.last_instruction, it.key());
	}
}

//------------------------------------------------------------------------------
// Name: is_thunk(edb::address_t address)
// Desc: basically returns true if the first instruction of the function is a 
//       jmp
//------------------------------------------------------------------------------
bool Analyzer::is_thunk(edb::address_t address) const {
	quint8 buf[edb::Instruction::MAX_SIZE];
	int buf_size = sizeof(buf);
	if(edb::v1::get_instruction_bytes(address, buf, buf_size)) {
		const edb::Instruction insn(buf, buf_size, address, std::nothrow);
		return insn.valid() && insn.type() == edb::Instruction::OP_JMP;
	}
	
	return false;
}

//------------------------------------------------------------------------------
// Name: 
// Desc: 
//------------------------------------------------------------------------------
void Analyzer::set_function_types_helper(Function &info) const {
	const edb::address_t addr = info.entry_address;

	if(is_thunk(addr)) {
		info.type = Function::FUNCTION_THUNK;
	} else {
		info.type = Function::FUNCTION_STANDARD;
	}
}

//------------------------------------------------------------------------------
// Name: set_function_types(FunctionMap &results)
// Desc: 
//------------------------------------------------------------------------------
void Analyzer::set_function_types(FunctionMap &results) {
	// give bonus if we have a symbol for the address
#ifdef USE_QT_CONCURRENT
	QtConcurrent::blockingMap(results, boost::bind(&Analyzer::set_function_types_helper, this, _1));
#else
	std::for_each(results.begin(), results.end(), boost::bind(&Analyzer::set_function_types_helper, this, _1));
#endif
}

//------------------------------------------------------------------------------
// Name: is_inside_known(const MemRegion &region, edb::address_t address)
// Desc: 
//------------------------------------------------------------------------------
bool Analyzer::is_inside_known(const MemRegion &region, edb::address_t address) {
	
	const FunctionMap &funcs = functions(region);
	Q_FOREACH(const Function &func, funcs) {
		if(address >= func.entry_address && address <= func.end_address) {
			return true;
		}
	}
	
	return false;
}

//------------------------------------------------------------------------------
// Name: 
// Desc: 
//------------------------------------------------------------------------------
void Analyzer::find_calls_from_known(const MemRegion &region, FunctionMap &results, QSet<edb::address_t> &walked_functions) {
	int updates;
	do {
		updates = walk_all_functions(results, region, walked_functions);
		qDebug() << "[Analyzer] got" << updates << "updates";
	} while(updates != 0);
}

//------------------------------------------------------------------------------
// Name: analyze(const MemRegion &region)
// Desc: 
//------------------------------------------------------------------------------
void Analyzer::analyze(const MemRegion &region_ref) {


	// NOTE: through a series of craziness, 
	// bonus_main, calls 
	// edb::v1::locate_main_function, which calls
	// edb::v1::primary_code_region, which calls
	// MemoryRegions::sync()
	// which happens to invalidate the references to regions we are potentially passed..
	// so just to be sure, we make a copy! wow, that was an annoying bug!
	MemRegion region(region_ref);

	RegionInfo &region_info = analysis_info_[region];

	QSettings settings;
	const bool fuzzy			= settings.value("Analyzer/fuzzy_logic_functions.enabled", true).toBool();
	const QByteArray md5		= md5_region(region);
	const QByteArray prev_md5	= region_info.md5;

	if(md5 != prev_md5 || fuzzy != region_info.fuzzy) {
		FunctionMap &function_map = region_info.analysis;

		function_map.clear();
				
		QSet<edb::address_t> walked_functions;

		// start with some obvious functions...
		emit update_progress(util::percentage(0, 12));
		qDebug("[Analyzer] adding entry points to the list...");
		bonus_entry_point(region, function_map);
		emit update_progress(util::percentage(1, 12));
		
		qDebug("[Analyzer] attempting to add 'main' to the list...");
		bonus_main(region, function_map);
		emit update_progress(util::percentage(2, 12));

		qDebug("[Analyzer] attempting to add functions with symbols to the list...");
		bonus_symbols(region, function_map);
		emit update_progress(util::percentage(3, 12));

		qDebug("[Analyzer] calculating function bounds... (pass 1)");
		find_calls_from_known(region, function_map, walked_functions);
		emit update_progress(util::percentage(4, 12));
		
		fix_overlaps(function_map);

		// ok, at this point, we've done the best we can with knowns
		// we should have a full analysis of all functions which are
		// reachable from known entry points in the code
		// let's start looking for unknowns
		if(fuzzy) {
			qDebug("[Analyzer] finding possible function calls...");

			// put these into a temporary bin
			FunctionMap found_functions;
			find_function_calls(region, found_functions);
			emit update_progress(util::percentage(6, 12));

			qDebug("[Analyzer] bonusing stack frames...");
			bonus_stack_frames(found_functions);
			emit update_progress(util::percentage(7, 12));

			// collect the ones which have enough references
			qDebug("[Analyzer] collecting high reference answers ...");
			for(FunctionMap::iterator it = found_functions.begin(); it != found_functions.end(); ) {
				if(it->reference_count >= MIN_REFCOUNT) {
					function_map[it->entry_address] = *it;
					found_functions.erase(it++);
				} else {
					++it;
				}
			}
			emit update_progress(util::percentage(8, 12));

			// walk the newly found functions!
			qDebug("[Analyzer] calculating function bounds... (pass 2)");
			find_calls_from_known(region, function_map, walked_functions);
			emit update_progress(util::percentage(9, 12));

			// ok, after this the only ones left should be low reference 
			// count ones...
			// but by now we can gaurantee they won't overlap with 
			// known good functions at all
			qDebug("[Analyzer] collecting low reference answers...");
			
			// promote weak symbols...
			SymbolManager &syms = edb::v1::symbol_manager();
			Q_FOREACH(const Function &func, found_functions) {
				if(!is_inside_known(region, func.entry_address)) {
					if(!function_map.contains(func.entry_address)) {
						function_map[func.entry_address] = func;
						
						const SymbolManager::SymbolP s = syms.find(func.entry_address);
						if(s && s->is_weak()) {
							function_map[func.entry_address].reference_count++;
						}
						
					}
				}
			}
			emit update_progress(util::percentage(10, 12));
			
			// walk the newly found functions!
			qDebug("[Analyzer] calculating function bounds... (pass 3)");
			find_calls_from_known(region, function_map, walked_functions);
			emit update_progress(util::percentage(11, 12));
		}
		
		qDebug("[Analyzer] determining function types...");
		set_function_types(function_map);
		emit update_progress(util::percentage(12, 12));

		qDebug("[Analyzer] complete");
		emit update_progress(100);

		region_info.md5		= md5;
		region_info.fuzzy	= fuzzy;
	} else {
		qDebug("[Analyzer] region unchanged, using previous analysis");
	}
}

//------------------------------------------------------------------------------
// Name: catagory(edb::address_t address) const
// Desc: 
//------------------------------------------------------------------------------
AnalyzerInterface::AddressCatagory Analyzer::catagory(edb::address_t address) const {

	Function func;
	if(find_containing_function(address, func)) {
		if(address == func.entry_address) {
			return ADDRESS_FUNC_START;
		} else if(address == func.end_address) {
			return ADDRESS_FUNC_END;
		} else {
			return ADDRESS_FUNC_BODY;
		}
	}
	return ADDRESS_FUNC_UNKNOWN;
}

//------------------------------------------------------------------------------
// Name: functions(const MemRegion &region) const
// Desc: 
//------------------------------------------------------------------------------
AnalyzerInterface::FunctionMap Analyzer::functions(const MemRegion &region) const {
	return analysis_info_[region].analysis;
}

//------------------------------------------------------------------------------
// Name: find_containing_function(edb::address_t address, AnalyzerInterface::Function &function) const
// Desc: 
//------------------------------------------------------------------------------
bool Analyzer::find_containing_function(edb::address_t address, AnalyzerInterface::Function &function) const {

	MemRegion region;
	if(edb::v1::memory_regions().find_region(address, region)) {
		const FunctionMap &funcs = functions(region);
		Q_FOREACH(const Function &f, funcs) {
			if(address >= f.entry_address && address <= f.end_address) {
				function = f;
				return true;
			}
		}
	}
	return false;
}

//------------------------------------------------------------------------------
// Name: md5_region(const MemRegion &region) const
// Desc: returns a byte array representing the MD5 of a region
//------------------------------------------------------------------------------
QByteArray Analyzer::md5_region(const MemRegion &region) const{

	static const edb::address_t page_size = edb::v1::debugger_core->page_size();

	const edb::address_t size_in_pages = region.size() / page_size;
	try {
		QVector<quint8> pages(size_in_pages * page_size);
	
		if(edb::v1::debugger_core->read_pages(region.start, &pages[0], size_in_pages)) {
			return edb::v1::get_md5(&pages[0], size_in_pages * page_size);
		}
		
	} catch(const std::bad_alloc &) {
		QMessageBox::information(0, tr("Memroy Allocation Error"),
			tr("Unable to satisfy memory allocation request for requested region."));
	}

	return QByteArray();
}

//------------------------------------------------------------------------------
// Name: 
// Desc: 
//------------------------------------------------------------------------------
edb::address_t Analyzer::module_entry_point(const MemRegion &region) const {
	edb::address_t entry = 0;
#ifdef Q_OS_LINUX
	const edb::address_t page_size = edb::v1::debugger_core->page_size();

	try {
		QVector<quint8> pages(page_size);
		if(edb::v1::debugger_core->read_pages(region.start, &pages[0], 1)) {

			union {
				void		*ptr;
				Elf32_Ehdr	*elf32;
				Elf64_Ehdr	*elf64;
			} code = { &pages[0] };

			if(std::memcmp(code.elf32->e_ident, ELFMAG, SELFMAG) == 0) {
				if(code.elf32->e_ident[EI_CLASS] == ELFCLASS32) {
					entry = code.elf32->e_entry;
				}
			}

			if(entry == 0 && std::memcmp(code.elf64->e_ident, ELFMAG, SELFMAG) == 0) {
				if(code.elf64->e_ident[EI_CLASS] == ELFCLASS64) {
					entry = code.elf64->e_entry;
				}
			}
		}
	} catch(const std::bad_alloc &) {
		QMessageBox::information(0, tr("Memroy Allocation Error"),
			tr("Unable to satisfy memory allocation request for requested region."));
	}
#else
	Q_UNUSED(region);
#endif
	return entry;
}


//------------------------------------------------------------------------------
// Name: bonus_entry_point(const MemRegion &region, FunctionMap &results) const
// Desc: 
//------------------------------------------------------------------------------
void Analyzer::bonus_entry_point(const MemRegion &region, FunctionMap &results) const {
	edb::address_t entry = module_entry_point(region);

	if(entry != 0) {
	
		// if the entry seems like a relative one (like for a library)
		// then add the base of its image
		if(entry < region.start) {
			entry += region.start;
		}
	
		// make sure we have an entry for this function
		if(region.contains(entry)) {
			update_results_entry(results, entry);	
		}
	}
}

//------------------------------------------------------------------------------
// Name: invalidate_analysis(const MemRegion &region)
// Desc: 
//------------------------------------------------------------------------------
void Analyzer::invalidate_analysis(const MemRegion &region) {
	analysis_info_[region] = RegionInfo();
}

//------------------------------------------------------------------------------
// Name: invalidate_analysis()
// Desc: 
//------------------------------------------------------------------------------
void Analyzer::invalidate_analysis() {
	analysis_info_.clear();
}

Q_EXPORT_PLUGIN2(Analyzer, Analyzer)
