/*
Copyright (C) 2006 - 2009 Evan Teran
                   eteran@alum.rit.edu

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
*/

#include "Bookmarks.h"
#include "Debugger.h"
#include "Expression.h"
#include "Debugger.h"

#include <QDockWidget>
#include <QGridLayout>
#include <QHeaderView>
#include <QInputDialog>
#include <QMainWindow>
#include <QMenu>
#include <QMessageBox>
#include <QPushButton>
#include <QShortcut>
#include <QSignalMapper>
#include <QTableWidget>

#include <memory>

//------------------------------------------------------------------------------
// Name: Bookmarks(
// Desc: 
//------------------------------------------------------------------------------
Bookmarks::Bookmarks() : menu_(0), button_add_(0), button_del_(0), button_clear_(0), table_(0) {
}

//------------------------------------------------------------------------------
// Name: 
// Desc: 
//------------------------------------------------------------------------------
void Bookmarks::add() {
	bool ok;
    const QString text = QInputDialog::getText(edb::v1::debugger_ui, tr("Bookmark Address"), tr("Address:"), QLineEdit::Normal, QString(), &ok);
	
	if (ok && !text.isEmpty()) {
		Expression<edb::address_t> expr(text, edb::v1::get_variable, edb::v1::get_value);
		ExpressionError err;

		const edb::address_t address = expr.evaluate_expression(ok, err);
		if(ok) {
			add_address(address);
		} else {
			QMessageBox::information(0, tr("Error In Address Expression!"), err.what());
		}
	}
}

//------------------------------------------------------------------------------
// Name: 
// Desc: 
//------------------------------------------------------------------------------
void Bookmarks::del() {
	
	std::auto_ptr<QTableWidgetItem> item(table_->takeItem(table_->currentRow(), 0));
	table_->removeRow(table_->currentRow());
	if(item.get()) {
		bool ok;
		const QString s = item->text();
		const edb::address_t address = edb::v1::string_to_address(s, ok);
		entries_.remove(address);
	}
}

//------------------------------------------------------------------------------
// Name: 
// Desc: 
//------------------------------------------------------------------------------
void Bookmarks::clear() {
	table_->clearContents();
	table_->setRowCount(0);
	entries_.clear();
}

//------------------------------------------------------------------------------
// Name: 
// Desc: 
//------------------------------------------------------------------------------
void Bookmarks::cellDoubleClicked(int row, int) {
	bool ok;
	QTableWidgetItem *const item = table_->item(row, 0);
	if(item != 0) {
		const QString s = item->text();
		const edb::address_t addr = edb::v1::string_to_address(s, ok);
		if(ok) {
			edb::v1::jump_to_address(addr);
		}
	}
}

//------------------------------------------------------------------------------
// Name: shortcut(int index)
// Desc: 
//------------------------------------------------------------------------------
void Bookmarks::shortcut(int index) {

	QTableWidgetItem *const item = table_->item(index, 0);
	
	if(item != 0) {
		emit cellDoubleClicked(index, 0);
	}
}

//------------------------------------------------------------------------------
// Name: menu(QWidget *parent)
// Desc: 
//------------------------------------------------------------------------------
QMenu *Bookmarks::menu(QWidget *parent) {
	if(menu_ == 0) {
	
		// if we are dealing with a main window (and we are...)
		// add the dock object
		QMainWindow *const mainWindow = qobject_cast<QMainWindow *>(parent);
		if(mainWindow != 0) {
			
			// boiler plate button stuff
			button_add_		= new QPushButton(tr("Add"));
			button_del_		= new QPushButton(tr("Del"));
			button_clear_	= new QPushButton(tr("Clear"));
			table_			= new QTableWidget;
			
			table_->horizontalHeader()->setStretchLastSection(true);
			table_->setEditTriggers(QAbstractItemView::NoEditTriggers);
			table_->setSelectionBehavior(QAbstractItemView::SelectRows);
			table_->verticalHeader()->hide();
			table_->setColumnCount(2);
			table_->setHorizontalHeaderItem(0, new QTableWidgetItem(tr("Address")));
			table_->setHorizontalHeaderItem(1, new QTableWidgetItem(tr("Comment")));
			table_->setContextMenuPolicy(Qt::CustomContextMenu);
			connect(table_, SIGNAL(customContextMenuRequested(const QPoint &)), SLOT(customContextMenuRequested(const QPoint &)));
			
			// events..
			connect(button_add_, SIGNAL(clicked()), this, SLOT(add()));
			connect(button_del_, SIGNAL(clicked()), this, SLOT(del()));
			connect(button_clear_, SIGNAL(clicked()), this, SLOT(clear()));
			connect(table_, SIGNAL(cellDoubleClicked(int,int)), this, SLOT(cellDoubleClicked(int,int)));

			// layouts
			QBoxLayout *const buttonLayout = new QBoxLayout(QBoxLayout::LeftToRight);
			buttonLayout->addWidget(button_add_);
			buttonLayout->addWidget(button_del_);
			buttonLayout->addWidget(button_clear_);

			QBoxLayout *const layout = new QBoxLayout(QBoxLayout::TopToBottom);
			layout->addWidget(table_);
			layout->addLayout(buttonLayout);

			QWidget *const widget = new QWidget(parent);
			widget->setLayout(layout);

			// make the dock widget and _name_ it, it is important to name it so
			// that it's state is saved in the GUI info
			QDockWidget *const dockWidget = new QDockWidget(tr("Bookmarks"), parent);
			dockWidget->setObjectName(QString::fromUtf8("Bookmarks"));
			dockWidget->setWidget(widget);

			// add it to the dock
			mainWindow->addDockWidget(Qt::RightDockWidgetArea, dockWidget);

			// make the menu and add the show/hide toggle for the widget
			menu_ = new QMenu(tr("Bookmarks"), parent);
			menu_->addAction(dockWidget->toggleViewAction());
			
			signal_mapper_ = new QSignalMapper(this);
			
			for(int i = 0; i < 10; ++i) {
				// create an action and attach it to the signal mapper
				QShortcut *const action = new QShortcut(QKeySequence(tr("Ctrl+%n", "", i)), mainWindow);
								
				if(i == 0) {
					signal_mapper_->setMapping(action, 9);
				} else {
					signal_mapper_->setMapping(action, i - 1);
				}
				
				connect(action, SIGNAL(activated()), signal_mapper_, SLOT(map()));
			}
			
			// connect the parametized signal to the slot..phew finally
			connect(signal_mapper_, SIGNAL(mapped(int)), this, SLOT(shortcut(int)));
		}
	}

	return menu_;
}

//------------------------------------------------------------------------------
// Name: 
// Desc: 
//------------------------------------------------------------------------------
QList<QAction *> Bookmarks::cpu_context_menu() {
	
	QList<QAction *> ret;
	
	QAction *const action_find = new QAction(tr("Add &Bookmark"), this);
	connect(action_find, SIGNAL(triggered()), this, SLOT(mnuBookmarkCPU()));
	ret << action_find;

	return ret;
}

//------------------------------------------------------------------------------
// Name: 
// Desc: 
//------------------------------------------------------------------------------
void Bookmarks::add_address(edb::address_t address) {
	if(!entries_.contains(address)) {
		QTableWidgetItem *const new_item = new QTableWidgetItem(edb::v1::format_pointer(address));
		const int row = table_->rowCount();
		table_->setRowCount(row + 1);
		table_->setItem(row, 0, new_item);
		table_->resizeColumnToContents(0);
		entries_.insert(address);
	}
}

//------------------------------------------------------------------------------
// Name: 
// Desc: 
//------------------------------------------------------------------------------
void Bookmarks::mnuBookmarkCPU() {
	add_address(edb::v1::cpu_selected_address());
}

//------------------------------------------------------------------------------
// Name: 
// Desc: 
//------------------------------------------------------------------------------
void Bookmarks::customContextMenuRequested(const QPoint &pos) {

	QMenu menu;
	QAction *const actionAdd		= menu.addAction(tr("&Add Address"));
	QAction *const actionDel		= menu.addAction(tr("&Delete Address"));
	QAction *const actionClear		= menu.addAction(tr("&Clear"));
	menu.addSeparator();
	QAction *const actionComment	= menu.addAction(tr("&Set Comment"));
	QAction *const chosen = menu.exec(table_->mapToGlobal(pos));

	if(chosen == actionAdd) {
		add();
	} else if(chosen == actionDel) {
		del();
	} else if(chosen == actionClear) {
		clear();
	} else if(chosen == actionComment) {
		bool ok;
		
		QString current_comment;
		QTableWidgetItem *const item = table_->item(table_->currentRow(), 1);
		if(item != 0) {
			current_comment = item->text();
		}
		
	    const QString text = QInputDialog::getText(table_, tr("Comment"), tr("Set Comment:"), QLineEdit::Normal, current_comment, &ok);
		if(ok) {
			table_->setItem(table_->currentRow(), 1, new QTableWidgetItem(text));
		}
	}
}

Q_EXPORT_PLUGIN2(Bookmarks, Bookmarks)
