/*
Copyright (C) 2006 - 2009 Evan Teran
                   eteran@alum.rit.edu

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
*/

#include "DialogMemoryRegions.h"
#include "DebuggerCoreInterface.h"
#include "Debugger.h"
#include "MemoryRegions.h"

#include <QHeaderView>
#include <QString>
#include <QMenu>
#include <QSortFilterProxyModel>
#include <QItemSelectionModel>

#include "ui_dialog_memoryregions.h"

//------------------------------------------------------------------------------
// Name: DialogMemoryRegions(QWidget *parent)
// Desc: 
//------------------------------------------------------------------------------
DialogMemoryRegions::DialogMemoryRegions(QWidget *parent) : QDialog(parent), ui(new Ui::DialogMemoryRegions) {
	ui->setupUi(this);

	ui->tableView->verticalHeader()->hide();
	ui->tableView->horizontalHeader()->setStretchLastSection(true);
	ui->tableView->horizontalHeader()->setResizeMode(QHeaderView::ResizeToContents);
	
	filter_model_ = new QSortFilterProxyModel(this);
	connect(ui->txtSearch, SIGNAL(textChanged(const QString &)), filter_model_, SLOT(setFilterFixedString(const QString &)));
}

//------------------------------------------------------------------------------
// Name: ~DialogMemoryRegions()
// Desc: 
//------------------------------------------------------------------------------
DialogMemoryRegions::~DialogMemoryRegions() {
	delete ui;
}

//------------------------------------------------------------------------------
// Name: showEvent(QShowEvent *)
// Desc: 
//------------------------------------------------------------------------------
void DialogMemoryRegions::showEvent(QShowEvent *) {

	filter_model_->setFilterKeyColumn(3);
	filter_model_->setSourceModel(&edb::v1::memory_regions());
	ui->tableView->setModel(filter_model_);
}

//------------------------------------------------------------------------------
// Name: on_tableWidget_customContextMenuRequested(const QPoint &pos)
// Desc: 
//------------------------------------------------------------------------------
void DialogMemoryRegions::on_tableView_customContextMenuRequested(const QPoint &pos) {
	
	QMenu menu;
	QMenu *const accessMenu = new QMenu(tr("Set Access"), this);
	accessMenu->addAction(tr("No Access"), this, SLOT(setAccessNone()));	// ---
	accessMenu->addAction(tr("Read Only"), this, SLOT(setAccessR()));		// r--
	accessMenu->addAction(tr("Write Only"), this, SLOT(setAccessW()));		// -w-
	accessMenu->addAction(tr("Execute Only"), this, SLOT(setAccessX()));	// --x
	accessMenu->addAction(tr("Read/Write"), this, SLOT(setAccessRW()));		// rw-
	accessMenu->addAction(tr("Read/Execute"), this, SLOT(setAccessRX()));	// r-x
	accessMenu->addAction(tr("Write/Execute"), this, SLOT(setAccessWX()));	// -wx
	accessMenu->addAction(tr("Full Access"), this, SLOT(setAccessRWX()));	// rwx
		
	
	menu.addMenu(accessMenu);
	menu.addSeparator();
	menu.addAction(tr("View in &CPU"), this, SLOT(viewInCPU()));
	menu.addAction(tr("View in &Stack"), this, SLOT(viewInStack()));
	menu.addAction(tr("View in &Dump"), this, SLOT(viewInDump()));
	menu.exec(ui->tableView->mapToGlobal(pos));
}

//------------------------------------------------------------------------------
// Name: selectedRegion() const
// Desc: 
//------------------------------------------------------------------------------
MemRegion *DialogMemoryRegions::selectedRegion() const {
	const QItemSelectionModel *const selModel = ui->tableView->selectionModel();
	const QModelIndexList sel = selModel->selectedRows();
	MemRegion *ret = 0;
	if(sel.size() == 1) {
		const QModelIndex index = filter_model_->mapToSource(sel[0]);
		ret = reinterpret_cast<MemRegion *>(index.internalPointer());
	}
	return ret;
}

//------------------------------------------------------------------------------
// Name: set_permissions(bool read, bool write, bool execute)
// Desc: 
//------------------------------------------------------------------------------
void DialogMemoryRegions::set_permissions(bool read, bool write, bool execute) {
	MemRegion *const region = selectedRegion();

	if(region != 0) {
		region->set_permissions(read, write, execute);
		edb::v1::memory_regions().sync();
	}
}

//------------------------------------------------------------------------------
// Name: setAccessNone()
// Desc: 
//------------------------------------------------------------------------------
void DialogMemoryRegions::setAccessNone() {
	set_permissions(false, false, false);
}

//------------------------------------------------------------------------------
// Name: setAccessR()
// Desc: 
//------------------------------------------------------------------------------
void DialogMemoryRegions::setAccessR() {
	set_permissions(true, false, false);
}

//------------------------------------------------------------------------------
// Name: setAccessW()
// Desc: 
//------------------------------------------------------------------------------
void DialogMemoryRegions::setAccessW() {
	set_permissions(false, true, false);
}

//------------------------------------------------------------------------------
// Name: setAccessX()
// Desc: 
//------------------------------------------------------------------------------
void DialogMemoryRegions::setAccessX() {
	set_permissions(false, false, true);
}

//------------------------------------------------------------------------------
// Name: setAccessRW()
// Desc: 
//------------------------------------------------------------------------------
void DialogMemoryRegions::setAccessRW() {
	set_permissions(true, true, false);
}

//------------------------------------------------------------------------------
// Name: setAccessRX()
// Desc: 
//------------------------------------------------------------------------------
void DialogMemoryRegions::setAccessRX() {
	set_permissions(true, false, true);
}

//------------------------------------------------------------------------------
// Name: setAccessWX()
// Desc: 
//------------------------------------------------------------------------------
void DialogMemoryRegions::setAccessWX() {
	set_permissions(false, true, true);
}

//------------------------------------------------------------------------------
// Name: setAccessRWX()
// Desc: 
//------------------------------------------------------------------------------
void DialogMemoryRegions::setAccessRWX() {
	set_permissions(true, true, true);
}

//------------------------------------------------------------------------------
// Name: 
// Desc: 
//------------------------------------------------------------------------------
void DialogMemoryRegions::viewInCPU() {
	MemRegion *const region = selectedRegion();

	if(region != 0) {
		edb::v1::jump_to_address(region->start);
	}
}

//------------------------------------------------------------------------------
// Name: 
// Desc: 
//------------------------------------------------------------------------------
void DialogMemoryRegions::viewInStack() {
	MemRegion *const region = selectedRegion();

	if(region != 0) {
		edb::v1::dump_stack(region->start, true);
	}
}

//------------------------------------------------------------------------------
// Name: 
// Desc: 
//------------------------------------------------------------------------------
void DialogMemoryRegions::viewInDump() {
	MemRegion *const region = selectedRegion();

	if(region != 0) {
		edb::v1::dump_data(region->start, true);
	}
}

