/*
Copyright (C) 2006 - 2009 Evan Teran
                   eteran@alum.rit.edu

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
*/

#include "Configuration.h"
#include "DebuggerCoreInterface.h"
#include "DebuggerPluginInterface.h"
#include "DebuggerMain.h"
#include "Debugger.h"
#include "DebuggerInternal.h"
#include "version.h"

#include <QApplication>
#include <QDateTime>
#include <QDir>
#include <QLibrary>
#include <QLibraryInfo>
#include <QMessageBox>
#include <QPluginLoader>
#include <QString>
#include <QTranslator>
#include <QtDebug>

#include <cstdlib>
#include <ctime>
#include <iostream>
#include <ostream>

#if defined(Q_OS_UNIX) && !defined(Q_OS_MACX)
#ifdef Q_OS_OPENBSD
#include <sys/exec_elf.h>
#else
#include <elf.h>
#endif
#include <fcntl.h>
#include <sys/mman.h>
#include <sys/stat.h>
#include <sys/types.h>
#endif

namespace {
#if defined(Q_OS_UNIX) && !defined(Q_OS_MACX)
	struct elf32_model {
		typedef quint32		address_t;

		typedef Elf32_Ehdr	Elf_header;
		typedef Elf32_Shdr	Elf_section_header;
		typedef Elf32_Sym	Elf_symbol;
		typedef Elf32_Rela	Elf_relocationA;
		typedef Elf32_Rel	Elf_relocation;

		static const int plt_entry_size = 0x10;

		static quint32 elf_r_sym(quint32 x)					{ return ELF32_R_SYM(x); }
		static quint32 elf_r_type(quint32 x)				{ return ELF32_R_TYPE(x); }
		static unsigned char elf_st_type(unsigned char x)	{ return ELF32_ST_TYPE(x); }
		static unsigned char elf_st_bind(unsigned char x)	{ return ELF32_ST_BIND(x); }

		struct symbol {
			address_t		address;
			unsigned int	size;
			QString			name;
			char			type;

			bool operator<(const symbol &rhs) const {
				return address < rhs.address || (address == rhs.address && name < rhs.name);
			}
			
			bool operator==(const symbol &rhs) const {
				return address == rhs.address && size == rhs.size && name == rhs.name && type == rhs.type;
			}

			QString to_string() const {
				char buf[256];
				qsnprintf(buf, sizeof(buf), "%08x %08x %c %s",
					address,
					size,
					type,
					qPrintable(name));
				return buf;
			}
		};
	};

	struct elf64_model {

		typedef quint64		address_t;

		typedef Elf64_Ehdr	Elf_header;
		typedef Elf64_Shdr	Elf_section_header;
		typedef Elf64_Sym	Elf_symbol;
		typedef Elf64_Rela	Elf_relocationA;
		typedef Elf64_Rel	Elf_relocation;
	
		static const int plt_entry_size = 0x10;

		static quint64 elf_r_sym(quint64 x)					{ return ELF64_R_SYM(x); }
		static quint64 elf_r_type(quint64 x)				{ return ELF64_R_TYPE(x); }
		static unsigned char elf_st_type(unsigned char x)	{ return ELF64_ST_TYPE(x); }
		static unsigned char elf_st_bind(unsigned char x)	{ return ELF64_ST_BIND(x); }

		struct symbol {
			address_t		address;
			unsigned int	size;
			QString			name;
			char			type;

			bool operator<(const symbol &rhs) const {
				return address < rhs.address || (address == rhs.address && name < rhs.name);
			}
			
			bool operator==(const symbol &rhs) const {
				return address == rhs.address && size == rhs.size && name == rhs.name && type == rhs.type;
			}

			QString to_string() const {
				char buf[256];
				qsnprintf(buf, sizeof(buf), "%016llx %08x %c %s",
					address,
					size,
					type,
					qPrintable(name));
				return buf;
			}
		};
	};

	bool is_elf32(void *ptr) {
		Elf32_Ehdr *const elf32_hdr = reinterpret_cast<Elf32_Ehdr *>(ptr);
		if(std::memcmp(elf32_hdr->e_ident, ELFMAG, SELFMAG) == 0) {
			return elf32_hdr->e_ident[EI_CLASS] == ELFCLASS32;
		}
		return false;
	}

	bool is_elf64(void *ptr) {
		Elf64_Ehdr *const elf64_hdr = reinterpret_cast<Elf64_Ehdr *>(ptr);
		if(std::memcmp(elf64_hdr->e_ident, ELFMAG, SELFMAG) == 0) {
			return elf64_hdr->e_ident[EI_CLASS] == ELFCLASS64;
		}
		return false;
	}


	/*
	   The  symbol  type.   At least the following types are used; others are, as well, depending on the object file format.  If lowercase,
	   the symbol is local; if uppercase, the symbol is global (external).

	   "A" The symbol's value is absolute, and will not be changed by further linking.

	   "B"
	   "b" The symbol is in the uninitialized data section (known as BSS).

	   "C" The symbol is common.  Common symbols are uninitialized data.  When linking, multiple common symbols may appear  with  the  same
    	   name.  If the symbol is defined anywhere, the common symbols are treated as undefined references.

	   "D"
	   "d" The symbol is in the initialized data section.

	   "G"
	   "g" The  symbol is in an initialized data section for small objects.  Some object file formats permit more efficient access to small
    	   data objects, such as a global int variable as opposed to a large global array.

	   "N" The symbol is a debugging symbol.

	   "p" The symbols is in a stack unwind section.

	   "R"
	   "r" The symbol is in a read only data section.

	   "S"
	   "s" The symbol is in an uninitialized data section for small objects.

	   "T"
	   "t" The symbol is in the text (code) section.

	   "U" The symbol is undefined.

	   "u" The  symbol  is  a unique global symbol.  This is a GNU extension to the standard set of ELF symbol bindings.  For such a symbol
    	   the dynamic linker will make sure that in the entire process_symbols there is just one symbol with this name and type in use.

	   "V"
	   "v" The symbol is a weak object.  When a weak defined symbol is linked with a normal defined symbol, the normal  defined  symbol  is
    	   used  with no error.  When a weak undefined symbol is linked and the symbol is not defined, the value of the weak symbol becomes
    	   zero with no error.  On some systems, uppercase indicates that a default value has been specified.

	   "W"
	   "w" The symbol is a weak symbol that has not been specifically tagged as a weak object symbol.  When a weak defined symbol is linked
    	   with  a  normal defined symbol, the normal defined symbol is used with no error.  When a weak undefined symbol is linked and the
    	   symbol is not defined, the value of the symbol is determined in a system-specific manner without error.  On some systems, upper-
    	   case indicates that a default value has been specified.

	   "-" The  symbol  is  a  stabs  symbol in an a.out object file.  In this case, the next values printed are the stabs other field, the
    	   stabs desc field, and the stab type.  Stabs symbols are used to hold debugging information.

	   "?" The symbol type is unknown, or object file format specific.
	*/

	template <class M>
	void process_symbols(const void *p, size_t size) {

		Q_UNUSED(size);
		
		typedef typename M::address_t			address_t;
		typedef typename M::Elf_header			Elf_header;
		typedef typename M::Elf_section_header	Elf_section_header;
		typedef typename M::Elf_symbol			Elf_symbol;
		typedef typename M::Elf_relocationA		Elf_relocationA;
		typedef typename M::Elf_relocation		Elf_relocation;
		typedef typename M::symbol				symbol;

		const uintptr_t base = reinterpret_cast<uintptr_t>(p);

		const Elf_header *const header					= static_cast<const Elf_header*>(p);
		const Elf_section_header *const sections_begin	= reinterpret_cast<Elf_section_header*>(base + header->e_shoff);
		const Elf_section_header *const sections_end	= sections_begin + header->e_shnum;
		const char *section_strings = reinterpret_cast<const char*>(base + sections_begin[header->e_shstrndx].sh_offset);


		address_t plt_address = 0;
		address_t got_address = 0;
		QSet<address_t>	plt_addresses;
		QList<symbol>	symbols;

		// collect special section addresses
		for(const Elf_section_header *section = sections_begin; section != sections_end; ++section) {
			if(strcmp(&section_strings[section->sh_name], ".plt") == 0) {
				plt_address = section->sh_addr;
			} else if(strcmp(&section_strings[section->sh_name], ".got") == 0) {
				got_address = section->sh_addr;
			}
		}

		// print out relocated symbols for special sections
		for(const Elf_section_header *section = sections_begin; section != sections_end; ++section) {
			address_t base_address = 0;
			if(strcmp(&section_strings[section->sh_name], ".rela.plt") == 0) {
				base_address = plt_address;
			} else if(strcmp(&section_strings[section->sh_name], ".rel.plt") == 0) {
				base_address = plt_address;		
			} else if(strcmp(&section_strings[section->sh_name], ".rela.got") == 0) {
				base_address = got_address;
			} else if(strcmp(&section_strings[section->sh_name], ".rel.got") == 0) {
				base_address = got_address;		
			} else {
				continue;
			}

			switch(section->sh_type) {
			case SHT_RELA:
				{
					int n = 0;
					const Elf_relocationA *relocation = reinterpret_cast<Elf_relocationA*>(base + section->sh_offset);

					for(size_t i = 0; i < section->sh_size / section->sh_entsize; ++i) {

						const int sym_index					= M::elf_r_sym(relocation[i].r_info);
						const Elf_section_header *linked	= &sections_begin[section->sh_link];
						const Elf_symbol *symbol_tab		= reinterpret_cast<Elf_symbol*>(base + linked->sh_offset);
						const char *string_tab				= reinterpret_cast<const char*>(base + sections_begin[linked->sh_link].sh_offset);

						const address_t symbol_address = base_address + ++n * M::plt_entry_size;

						const char *sym_name = &section_strings[section->sh_name];
						if(strlen(sym_name) > (sizeof(".rela.") - 1) && memcmp(sym_name, ".rela.", (sizeof(".rela.") - 1)) == 0) {
							sym_name += 6;
						}

						plt_addresses.insert(symbol_address);

						symbol sym;
						sym.address	= symbol_address;
						sym.size	= (symbol_tab[sym_index].st_size ? symbol_tab[sym_index].st_size : 0x10);
						sym.name	= &string_tab[symbol_tab[sym_index].st_name];
						sym.name	+= "@";
						sym.name	+= sym_name;
						sym.type	= 'P';
						symbols.push_back(sym);						
					}

				}
				break;
			case SHT_REL:
				{
					int n = 0;
					const Elf_relocation *relocation = reinterpret_cast<Elf_relocation*>(base + section->sh_offset);

					for(size_t i = 0; i < section->sh_size / section->sh_entsize; ++i) {

						const int sym_index					= M::elf_r_sym(relocation[i].r_info);
						const Elf_section_header *linked	= &sections_begin[section->sh_link];
						const Elf_symbol *symbol_tab		= reinterpret_cast<Elf_symbol*>(base + linked->sh_offset);
						const char *string_tab				= reinterpret_cast<const char*>(base + sections_begin[linked->sh_link].sh_offset);

						const address_t symbol_address = base_address + ++n * M::plt_entry_size;

						const char *sym_name = &section_strings[section->sh_name];
						if(strlen(sym_name) > (sizeof(".rel.") - 1) && memcmp(sym_name, ".rel.", (sizeof(".rel.") - 1)) == 0) {
							sym_name += 5;
						}

						plt_addresses.insert(symbol_address);

						symbol sym;
						sym.address	= symbol_address;
						sym.size	= (symbol_tab[sym_index].st_size ? symbol_tab[sym_index].st_size : 0x10);
						sym.name	= &string_tab[symbol_tab[sym_index].st_name];
						sym.name	+= "@";
						sym.name	+= sym_name;
						sym.type	= 'P';
						symbols.push_back(sym);						
					}

				}
				break;
			}
		}

		// collect regular symbols
		for(const Elf_section_header *section = sections_begin; section != sections_end; ++section) {

			switch(section->sh_type) {			
			case SHT_SYMTAB:
			case SHT_DYNSYM:	
				{
					Elf_symbol *symbol_tab	= reinterpret_cast<Elf_symbol*>(base + section->sh_offset);
					const char *string_tab	= reinterpret_cast<const char*>(base + sections_begin[section->sh_link].sh_offset);

					for(size_t i = 0; i < section->sh_size / section->sh_entsize; ++i) {

						const Elf_section_header *related_section	= 0;

						if(symbol_tab[i].st_shndx != SHN_UNDEF && symbol_tab[i].st_shndx < SHN_LORESERVE) {
							related_section = &sections_begin[symbol_tab[i].st_shndx];
						}


						if(plt_addresses.find(symbol_tab[i].st_value) == plt_addresses.end()) {

							if(symbol_tab[i].st_value && strlen(&string_tab[symbol_tab[i].st_name]) > 0) {

								symbol sym;
								sym.address	= symbol_tab[i].st_value;
								sym.size	= symbol_tab[i].st_size;
								sym.name	= &string_tab[symbol_tab[i].st_name];
								sym.type	= (M::elf_st_type(symbol_tab[i].st_info) == STT_FUNC ? 'T' : 'D');
								symbols.push_back(sym);
							}
						}
					}
				}
				break;
			}
		}

		// collect unnamed symbols
		for(const Elf_section_header *section = sections_begin; section != sections_end; ++section) {

			switch(section->sh_type) {			
			case SHT_SYMTAB:
			case SHT_DYNSYM:	
				{
					Elf_symbol *symbol_tab	= reinterpret_cast<Elf_symbol*>(base + section->sh_offset);
					const char *string_tab	= reinterpret_cast<const char*>(base + sections_begin[section->sh_link].sh_offset);

					for(size_t i = 0; i < section->sh_size / section->sh_entsize; ++i) {

						const Elf_section_header *related_section	= 0;

						if(symbol_tab[i].st_shndx != SHN_UNDEF && symbol_tab[i].st_shndx < SHN_LORESERVE) {
							related_section = &sections_begin[symbol_tab[i].st_shndx];
						}

						if(plt_addresses.find(symbol_tab[i].st_value) == plt_addresses.end()) {

							if(symbol_tab[i].st_value && strlen(&string_tab[symbol_tab[i].st_name]) == 0) {
								symbol sym;
								sym.address	= symbol_tab[i].st_value;
								sym.size	= symbol_tab[i].st_size;
								sym.name	= QString("$sym_%1").arg(edb::v1::format_pointer(symbol_tab[i].st_value));
								sym.type	= (M::elf_st_type(symbol_tab[i].st_info) == STT_FUNC ? 'T' : 'D');
								symbols.push_back(sym);
							}
						}
					}
				}
				break;
			}
		}

		qSort(symbols.begin(), symbols.end());
		typename QList<symbol>::const_iterator new_end = std::unique(symbols.begin(), symbols.end());
		for(typename QList<symbol>::const_iterator it = symbols.begin(); it != new_end; ++it) {
			std::cout << qPrintable(it->to_string()) << '\n';
		}
	}
#endif

	//--------------------------------------------------------------------------
	// Name: 
	// Desc: 
	//--------------------------------------------------------------------------
	void generate_symbols(const QString &filename) {
        #if defined(Q_OS_UNIX) && !defined(Q_OS_MACX)
		QFile file(filename);
		if(file.open(QIODevice::ReadOnly)) {			
			std::cout << qPrintable(QDateTime::currentDateTime().toString("ddd MMM d hh:mm:ss yyyy")) << '\n';

			QByteArray md5 = edb::v1::get_file_md5(filename);
			std::cout << md5.toHex().data() << ' ' << qPrintable(QFileInfo(filename).absoluteFilePath()) << '\n';

			void *const file_ptr = reinterpret_cast<void *>(file.map(0, file.size(), QFile::NoOptions));
			if(file_ptr) {
				if(is_elf64(file_ptr)) {
					process_symbols<elf64_model>(file_ptr, file.size());
				} else if(is_elf32(file_ptr)) {
					process_symbols<elf32_model>(file_ptr, file.size());
				}
			}
		}
	#else
		std::cerr << "generating symbols is not yet supported on non-unix targets" << std::endl;
	#endif
	}


	//--------------------------------------------------------------------------
	// Name: load_plugins(const QString &directory)
	// Desc: attemps to load all plugins in a given directory
	//--------------------------------------------------------------------------
	void load_plugins(const QString &directory) {

		QDir plugins_dir(qApp->applicationDirPath());

		// TODO: attempt to detect the same plugin being loaded twice

		plugins_dir.cd(directory);

		Q_FOREACH(const QString &file_name, plugins_dir.entryList(QDir::Files)) {
			if(QLibrary::isLibrary(file_name)) {
				const QString full_path = plugins_dir.absoluteFilePath(file_name);
				QPluginLoader loader(full_path);
				loader.setLoadHints(QLibrary::ExportExternalSymbolsHint);

				QObject *const plugin = loader.instance();

				if (plugin != 0) {

					// TODO: handle the case where we find more than one core plugin...
					DebuggerCoreInterface *const core_plugin = qobject_cast<DebuggerCoreInterface *>(plugin);
					if(core_plugin != 0) {
						if(edb::v1::debugger_core == 0) {
							qDebug("[load_plugins] Loading Core Plugin: %-25s : %p", qPrintable(file_name), static_cast<void *>(plugin));
							edb::v1::debugger_core = core_plugin;
						}
					}

					DebuggerPluginInterface *const generic_plugin = qobject_cast<DebuggerPluginInterface *>(plugin);
					if(generic_plugin != 0) {
						if(edb::internal::register_plugin(full_path, plugin)) {
							// we have a general purpose plugin
							qDebug("[load_plugins] Loading Plugin: %-30s : %p", qPrintable(file_name), static_cast<void *>(plugin));
						}
					}
				} else {
					qDebug() << "[load_plugins]" << qPrintable(loader.errorString());
				}
			}
		}
	}

	//--------------------------------------------------------------------------
	// Name: start_debugger(edb::pid_t attach_pid)
	// Desc: starts the main debugger code
	//--------------------------------------------------------------------------
	int start_debugger(edb::pid_t attach_pid, const QString &program, const QStringList &programArgs) {

		qDebug() << "Starting EDB Version:" << version_string;
		qDebug("Please Report Bugs & Requests At: http://bugs.codef00.com/");

		// create the main window object	
		DebuggerMain debugger;

		// look for some plugins..
		load_plugins(QDir().absolutePath());
		
		const QString plugin_path = edb::v1::config().plugin_path;
		if(QDir().absolutePath() != plugin_path) {
			load_plugins(plugin_path);
		}

		// let the plugins setup their menus
		debugger.finish_plugin_setup(edb::v1::plugin_list());

		// ok things are initialized to a reasonable degree, let's show the main window
		debugger.show();

		// have we been asked to attach to a given program?	
		if(attach_pid != 0) {
			debugger.attach(attach_pid);
		} else if(!program.isEmpty()) {
			debugger.execute(program, programArgs);
		}

		if(edb::v1::debugger_core == 0) {
			QMessageBox::warning(0, 
				QObject::tr("EDB Failed To Load A Neccessary Plugin"), 
				QObject::tr(
				"Failed to successfully load the debugger core plugin. Please make sure it exists and that the plugin path is correctly configured.\n"
				"This is normal if EDB has not been previously run or the configuration file has been removed."));

			edb::v1::dialog_options()->exec();
			
			QMessageBox::warning(0,
				QObject::tr("EDB"), 
				QObject::tr("EDB will now close. If you were successful in specifying the location of the debugger core plugin, please run EDB again.")
				);

			// TODO: detect if they corrected the issue and try again
			return -1;
		} else {
	    	return qApp->exec();
		}
	}
}


//------------------------------------------------------------------------------
// Name: main(int argc, char *argv[])
// Desc: entry point
//------------------------------------------------------------------------------
int main(int argc, char *argv[]) {
	
	QT_REQUIRE_VERSION(argc, argv, "4.5.0");

	QApplication app(argc, argv);
	QApplication::setWindowIcon(QIcon(":/debugger/images/edb48-logo.png"));
	
	std::srand(std::time(0));
	
	// setup organization info so settings go in right place
	QApplication::setOrganizationName("codef00.com");
	QApplication::setOrganizationDomain("codef00.com");
	QApplication::setApplicationName("edb");
	
	// load some translations
	QTranslator qtTranslator;
	qtTranslator.load("qt_" + QLocale::system().name(), QLibraryInfo::location(QLibraryInfo::TranslationsPath));
	app.installTranslator(&qtTranslator);

	QTranslator myappTranslator;
	myappTranslator.load("edb_" + QLocale::system().name());
	app.installTranslator(&myappTranslator);
	
	const QStringList args = app.arguments();
	edb::pid_t	attach_pid = 0;
	QStringList	run_args;
	QString		run_app;

	if(args.size() > 1) {
		if(args.size() == 3 && args[1] == "--attach") {
			attach_pid = args[2].toUInt();
		} else if(args.size() >= 3 && args[1] == "--run") {
			run_app = args[2];
			
			for(int i = 3; i < args.size(); ++i) {
				run_args.push_back(args[i]);
			}
		} else if(args.size() == 3 && args[1] == "--symbols") {
			generate_symbols(args[2]);
			return 0;
		} else if(args.size() == 2 && args[1] == "--version") {
			std::cout << "EDB Version: " << version_string << std::endl;
			return 0;
		} else {
			std::cerr << "usage: " << qPrintable(args[0]) << " [--symbols <filename>] [ --attach <pid> ] [ --run <program> (args...) ] [ --version ]" << std::endl;
			return -1;
		}
	}

	return start_debugger(attach_pid, run_app, run_args);
}
