/*
Copyright (C) 2006 - 2009 Evan Teran
                   eteran@alum.rit.edu

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
*/

#include "QDisassemblyView.h"
#include "AnalyzerInterface.h"
#include "ArchProcessorInterface.h"
#include "Configuration.h"
#include "Debugger.h"
#include "DebuggerCoreInterface.h"
#include "Instruction.h"
#include "SymbolManager.h"
#include "SyntaxHighlighter.h"
#include "Util.h"
#include "format_address.h"

#include <QAbstractItemDelegate>
#include <QAbstractTextDocumentLayout>
#include <QMouseEvent>
#include <QPainter>
#include <QPixmap>
#include <QScrollBar>
#include <QTextDocument>
#include <QTextDocument>
#include <QTextLayout>
#include <QToolTip>
#include <QtGlobal>
#include <climits>

#define DEFAULT_BYTE_WIDTH 8

void draw_rich_text(QPainter *painter, int x, int y, const QTextDocument &text) {
	painter->save();
	painter->translate(x, y);
	QAbstractTextDocumentLayout::PaintContext context;
	context.palette.setColor(QPalette::Text, painter->pen().color());
	text.documentLayout()->draw(painter, context);
	painter->restore();
}

//------------------------------------------------------------------------------
// Name: QDisassemblyView(QWidget * parent)
// Desc: constructor
//------------------------------------------------------------------------------
QDisassemblyView::QDisassemblyView(QWidget * parent) : QAbstractScrollArea(parent),
		address_offset_(0), current_address_(0), 
		breakpoint_icon_(":/debugger/images/edb14-breakpoint.png"),
		current_address_icon_(":/debugger/images/edb14-arrow.png"),
		selected_instruction_address_(0), 
		selected_instruction_size_(0), moving_line1_(false),
		moving_line2_(false), line1_(0), line2_(0),
		highlighter_(new SyntaxHighlighter(this)) {
		

	setShowAddressSeparator(true);
	
	setFont(QFont("Monospace", 8));
	setMouseTracking(true);
	setVerticalScrollBarPolicy(Qt::ScrollBarAlwaysOn);
}

//------------------------------------------------------------------------------
// Name: 
// Desc: 
//------------------------------------------------------------------------------
QDisassemblyView::~QDisassemblyView() {
}

//------------------------------------------------------------------------------
// Name: setShowAddressSeparator(bool value)
// Desc: 
//------------------------------------------------------------------------------
void QDisassemblyView::setShowAddressSeparator(bool value) {
	util::format_address<edb::address_t>::format_string(address_format_string_, value);
	show_address_separator_ = value;
}

//------------------------------------------------------------------------------
// Name: formatAddress(edb::address_t address)
// Desc: 
//------------------------------------------------------------------------------
QString QDisassemblyView::formatAddress(edb::address_t address) const {
	return util::format_address<edb::address_t>::format(address_format_string_, address);
}

//------------------------------------------------------------------------------
// Name: repaint()
// Desc: 
//------------------------------------------------------------------------------
void QDisassemblyView::repaint() {
	viewport()->repaint();
}

//------------------------------------------------------------------------------
// Name: addressShown(edb::address_t address) const
// Desc: returns true if a given address is in the visible range
//------------------------------------------------------------------------------
bool QDisassemblyView::addressShown(edb::address_t address) const {
	return show_addresses_.contains(address);
}

//------------------------------------------------------------------------------
// Name: setCurrentAddress(edb::address_t address)
// Desc: sets the 'current address' (where EIP is usually)
//------------------------------------------------------------------------------
void QDisassemblyView::setCurrentAddress(edb::address_t address) {
	current_address_ = address;
}

//------------------------------------------------------------------------------
// Name: setRegion(const MemRegion &r)
// Desc: sets the memory region we are viewing
//------------------------------------------------------------------------------
void QDisassemblyView::setRegion(const MemRegion &r) {
	if(region_ != r) {
		region_ = r;
		updateScrollbars();
	}
	repaint();
}

//------------------------------------------------------------------------------
// Name: clear()
// Desc: clears the display
//------------------------------------------------------------------------------
void QDisassemblyView::clear() {
	setRegion(MemRegion());
}

//------------------------------------------------------------------------------
// Name: setAddressOffset(edb::address_t address)
// Desc: 
//------------------------------------------------------------------------------
void QDisassemblyView::setAddressOffset(edb::address_t address) {
	address_offset_ = address;
}

//------------------------------------------------------------------------------
// Name: scrollTo(edb::address_t address)
// Desc: 
//------------------------------------------------------------------------------
void QDisassemblyView::scrollTo(edb::address_t address) {
	verticalScrollBar()->setValue(address - address_offset_);
}

//------------------------------------------------------------------------------
// Name: format_instruction_bytes(const edb::Instruction &insn, int maxStringPx, const QFontMetrics &metrics) const
// Desc: 
//------------------------------------------------------------------------------
QString QDisassemblyView::format_instruction_bytes(const edb::Instruction &insn, int maxStringPx, const QFontMetrics &metrics) const {
	const QString byteBuffer = edb::v1::format_bytes(QByteArray::fromRawData(reinterpret_cast<const char *>(insn.buffer()), insn.size()));
	return metrics.elidedText(byteBuffer, Qt::ElideRight, maxStringPx);
}

//------------------------------------------------------------------------------
// Name: format_instruction_bytes(const edb::Instruction &insn) const
// Desc: 
//------------------------------------------------------------------------------
QString QDisassemblyView::format_instruction_bytes(const edb::Instruction &insn) const {
	return edb::v1::format_bytes(QByteArray::fromRawData(reinterpret_cast<const char *>(insn.buffer()), insn.size()));
}

static const QColor filling_dis_colour	= Qt::gray;
static const QColor default_dis_colour	= Qt::blue;
static const QColor invalid_dis_colour	= Qt::blue;
static const QColor data_dis_colour		= Qt::blue;

//------------------------------------------------------------------------------
// Name: 
// Desc: 
//------------------------------------------------------------------------------
int QDisassemblyView::draw_instruction(QPainter &painter, const edb::Instruction &insn, bool upper, int y, int line_height, int l2) const {

	const bool is_filling	= edb::v1::arch_processor().is_filling(insn);
	int x					= font_width_ + font_width_ + l2 + (font_width_ / 2);
	const int ret			= insn.size();
	
	if(insn.valid()) {
		QString opcode = QString::fromStdString(insn.format(upper));
		
		if(is_filling) {
			painter.setPen(filling_dis_colour);
			painter.drawText(
				x,
				y, 
				opcode.length() * font_width_, 
				line_height, 
				Qt::AlignTop, 
				opcode);
		} else {
		
			switch(insn.type()) {
			case edb::Instruction::OP_JCC:
			case edb::Instruction::OP_JMP:
			case edb::Instruction::OP_LOOP:
			case edb::Instruction::OP_LOOPE:
			case edb::Instruction::OP_LOOPNE:
			case edb::Instruction::OP_CALL:
				if (insn.operand_count() != 0) {
					const edb::Operand &oper = insn.operand(0);
					if(oper.general_type() == edb::Operand::TYPE_REL) {
						const edb::Instruction::address_t target	= oper.relative_target();
						const SymbolManager::SymbolP sym			= edb::v1::symbol_manager().find(target);
						if(sym != 0) {
							opcode.append(QString(" <%2>").arg(sym->name));
						}
					}
				}
				break;
			default:
				break;
			}
		
			painter.setPen(default_dis_colour);
			QTextDocument doc;
			doc.setDefaultFont(font());
			doc.setDocumentMargin(0);
			doc.setPlainText(opcode);
			highlighter_->setDocument(&doc);
			draw_rich_text(&painter, x, y, doc);
		}

	} else {
		const QString asm_buffer = format_invalid_instruction_bytes(insn, painter);
		
		painter.drawText(
			x,
			y, 
			asm_buffer.length() * font_width_, 
			line_height, 
			Qt::AlignTop, 
			asm_buffer);
	}
	
	return ret;
}

//------------------------------------------------------------------------------
// Name: 
// Desc: 
//------------------------------------------------------------------------------
QString QDisassemblyView::format_invalid_instruction_bytes(const edb::Instruction &insn, QPainter &painter) const {
	char byte_buffer[32];
	const quint8 *const buf = insn.buffer();

	switch(insn.size()) {
	case 1:
		painter.setPen(data_dis_colour);
		qsnprintf(byte_buffer, sizeof(byte_buffer), "db 0x%02x", *buf & 0xff);
		break;
	case 2:
		painter.setPen(data_dis_colour);
		qsnprintf(byte_buffer, sizeof(byte_buffer), "dw 0x%04x", *reinterpret_cast<const quint16 *>(buf) & 0xffff);
		break;
	case 4:
		painter.setPen(data_dis_colour);
		qsnprintf(byte_buffer, sizeof(byte_buffer), "dd 0x%08x", *reinterpret_cast<const quint32 *>(buf) & 0xffffffff);
		break;
	case 8:
		painter.setPen(data_dis_colour);
		qsnprintf(byte_buffer, sizeof(byte_buffer), "dq 0x%016llx", *reinterpret_cast<const quint64 *>(buf));
		break;
	default:
		// we tried...didn't we?
		painter.setPen(invalid_dis_colour);
		return tr("invalid");
	}
	return byte_buffer;
}

//------------------------------------------------------------------------------
// Name: 
// Desc: 
//------------------------------------------------------------------------------
void QDisassemblyView::drawFunctionMarkers(QPainter &painter, edb::address_t address, int l2, int y, int insn_size, AnalyzerInterface *analyzer) {
	Q_CHECK_PTR(analyzer);
	painter.setPen(QPen(palette().shadow().color(), 2));

	const AnalyzerInterface::AddressCatagory cat = analyzer->catagory(address);
	const int line_height = this->line_height();
	const int x = l2 + font_width_;

	switch(cat) {
	case AnalyzerInterface::ADDRESS_FUNC_START:

		// half of a horizontal
		painter.drawLine(
			x,
			y + line_height / 2,
			l2 + (font_width_ / 2) + font_width_,
			y + line_height / 2
			);

		// half of a vertical
		painter.drawLine(
			x, 
			y + line_height / 2, 
			x, 
			y + line_height
			);

		break;
	case AnalyzerInterface::ADDRESS_FUNC_BODY:
		if(analyzer->catagory(address + insn_size - 1) == AnalyzerInterface::ADDRESS_FUNC_END) {
			goto do_end;
		} else {

			// vertical line
			painter.drawLine(
				x, 
				y, 
				x, 
				y + line_height);

		}
		break;
	case AnalyzerInterface::ADDRESS_FUNC_END:
	do_end:

		// half of a vertical
		painter.drawLine(
			x, 
			y, 
			x, 
			y + line_height / 2
			);

		// half of a horizontal
		painter.drawLine(
			x,
			y + line_height / 2,
			l2 + (font_width_ / 2) + font_width_,
			y + line_height / 2
			);
	default:
		break;
	}
}

//------------------------------------------------------------------------------
// Name: paintEvent(QPaintEvent *)
// Desc: 
//------------------------------------------------------------------------------
void QDisassemblyView::paintEvent(QPaintEvent *) {

	QPainter painter(viewport());
	
	const bool uppercase = edb::v1::config().uppercase_disassembly;
		
	int viewable_lines		= viewport()->height() / line_height();
	int current_line			= verticalScrollBar()->value();
	int row_index			= 0;
	int y					= 0;
	const int l1			= line1();
	const int l2			= line2();
	const int line_height	= this->line_height();
	
	// TODO: reimplement me 
	// const Configuration::Syntax syntax = edb::v1::config().syntax;
	const int regionSize = region_.size();
	
	if(regionSize == 0) {
		return;
	}
	
	show_addresses_.clear();
	show_addresses_.insert(address_offset_ + current_line);
	
	const int bytesWidth = l2 - l1;
	
	const QBrush alternatedBaseColour	= palette().alternateBase();
	const QBrush bytesColour			= palette().text();
	const QBrush dividerColour			= palette().shadow();
	const QPen bytesPen					= bytesColour.color();
	const QPen dividerPen				= dividerColour.color();
	
	AnalyzerInterface *const analyzer = edb::v1::analyzer();
	
	edb::address_t last_address = 0;

	while(viewable_lines >= 0 && current_line < regionSize) {
		const edb::address_t address = address_offset_ + current_line;

		quint8 buf[edb::Instruction::MAX_SIZE + 1];

		// do the longest read we can while still not passing the region end
		int buf_size = qMin<edb::address_t>((region_.end - address), sizeof(buf));

		// read in the bytes...
		if(!edb::v1::get_instruction_bytes(address, buf, buf_size)) {
			// if the read failed, let's pretend that we were able to read a 
			// single 0xff byte so that we have _something_ to display.
			buf_size = 1;
			*buf = 0xff;
		}

		// disassemble the instruction
		const edb::Instruction insn1(buf, buf_size, address, std::nothrow);
		const edb::Instruction insn2(buf, 1, address, std::nothrow);
		
		// has the analyzer decided that the next instruction is a block start? if
		// so, then cut this one short so that it doesn't overlap with the block
		// anymore
		const edb::Instruction &insn(
			(analyzer != 0) && (analyzer->catagory(address + 1) == AnalyzerInterface::ADDRESS_FUNC_START) ?
			insn2 :
			insn1);

		const int insn_size = insn.size();
		
		//TODO: remove this hack here
		if(insn_size == 0) {
			return;
		}
				
		// draw alternating line backgrounds
		if(row_index & 1) {
			painter.fillRect(0, y, width(), line_height, alternatedBaseColour);
		}
		
		if(analyzer) {
			drawFunctionMarkers(painter, address, l2, y, insn_size, analyzer);
		}

		// draw breakpoint icon or eip indicator
		if(address == current_address_) {
			painter.drawPixmap(1, y + 1, current_address_icon_);
		} else if(edb::v1::find_breakpoint(address) != 0) {
			painter.drawPixmap(1, y + 1, breakpoint_icon_);
		}
		
		// format the different components
		const QString byteBuffer	= format_instruction_bytes(insn, bytesWidth, painter.fontMetrics());
		const QString addressBuffer	= formatAddress(address);

		// draw the address
		painter.setPen(QPen(Qt::red));
		painter.drawText(
			breakpoint_icon_.width() + 1, 
			y, 
			addressBuffer.length() * font_width_, 
			line_height, 
			Qt::AlignTop, 
			addressBuffer);

		// draw the data bytes
		painter.setPen(bytesPen);
		painter.drawText(
			l1 + (font_width_ / 2),
			y, 
			byteBuffer.length() * font_width_, 
			line_height, 
			Qt::AlignTop, 
			byteBuffer);


		//optionally draw the symbol name
		const SymbolManager::SymbolP sym = edb::v1::symbol_manager().find(address);
		if(sym) {

			const int maxStringPx = l1 - (breakpoint_icon_.width() + 1 + ((addressBuffer.length() + 1) * font_width_));

			if(maxStringPx >= font_width_) {
				const QString symBuffer = painter.fontMetrics().elidedText(sym->name, Qt::ElideRight, maxStringPx);

				painter.drawText(
					breakpoint_icon_.width() + 1 + ((addressBuffer.length() + 1) * font_width_), 
					y, 
					symBuffer.length() * font_width_, 
					line_height, 
					Qt::AlignTop, 
					symBuffer);
			}
		}

		// for relative jumps draw the jump direction indicators
		switch(insn.type()) {
		case edb::Instruction::OP_JCC:
		case edb::Instruction::OP_JMP:
		case edb::Instruction::OP_LOOP:
		case edb::Instruction::OP_LOOPE:
		case edb::Instruction::OP_LOOPNE:
			if(insn.operand(0).general_type() == edb::Operand::TYPE_REL) {
				const edb::Instruction::address_t target = insn.operand(0).relative_target();
			
				painter.drawText(
					l2 + font_width_ + (font_width_ / 2),
					y, 
					font_width_, 
					line_height, 
					Qt::AlignTop,
					QString((target > address) ? QChar(0x02C7) : QChar(0x02C6))
					);
			}
			break;
		default:
			break;
		}
		
		// draw the disassembly
		current_line += draw_instruction(painter, insn, uppercase, y, line_height, l2);
		show_addresses_.insert(address);
		last_address = address;
			
		y += line_height;
		++row_index;
		--viewable_lines;
	}
	
	show_addresses_.remove(last_address);
	
	painter.setPen(dividerPen);
	painter.drawLine(l1, 0, l1, height());
	painter.drawLine(l2, 0, l2, height());
}

//------------------------------------------------------------------------------
// Name: setFont(const QFont &f)
// Desc: overloaded version of setFont, calculates font metrics for later
//------------------------------------------------------------------------------
void QDisassemblyView::setFont(const QFont &f) {

	// TODO: assert that we are using a fixed font & find out if we care?
	QAbstractScrollArea::setFont(f);

	// recalculate all of our metrics/offsets
	const QFontMetrics metrics(f);
	font_width_		= metrics.width('X');
	font_height_	= metrics.height();

	updateScrollbars();
}

//------------------------------------------------------------------------------
// Name: resizeEvent(QResizeEvent *)
// Desc: 
//------------------------------------------------------------------------------
void QDisassemblyView::resizeEvent(QResizeEvent *) {
	updateScrollbars();
}

//------------------------------------------------------------------------------
// Name: line_height() const
// Desc: 
//------------------------------------------------------------------------------
int QDisassemblyView::line_height() const {
	return qMax(font_height_, current_address_icon_.height());
}

//------------------------------------------------------------------------------
// Name: updateScrollbars()
// Desc: 
//------------------------------------------------------------------------------
void QDisassemblyView::updateScrollbars() {
	const unsigned int total_lines		= region_.size();
	const unsigned int viewable_lines	= viewport()->height() / line_height();
	const unsigned int scroll_max 		= (total_lines > viewable_lines) ? total_lines - 1 : 0;
		
	verticalScrollBar()->setMaximum(scroll_max);
}

//------------------------------------------------------------------------------
// Name: auto_line1() const
// Desc: 
//------------------------------------------------------------------------------
int QDisassemblyView::auto_line1() const {
	const unsigned int elements = address_length();
	return (elements * font_width_) + (font_width_ / 2) + breakpoint_icon_.width() + 1;
}

//------------------------------------------------------------------------------
// Name: line1() const
// Desc: 
//------------------------------------------------------------------------------
int QDisassemblyView::line1() const {
	if(line1_ == 0) {
		return auto_line1();
	} else {
		return line1_;
	}
}

//------------------------------------------------------------------------------
// Name: line2() const
// Desc: 
//------------------------------------------------------------------------------
int QDisassemblyView::line2() const {
	if(line2_ == 0) {
		return line1() + (DEFAULT_BYTE_WIDTH * 3) * font_width_;
	} else {
		return line2_;
	}
}

//------------------------------------------------------------------------------
// Name: address_length() const
// Desc: 
//------------------------------------------------------------------------------
int QDisassemblyView::address_length() const {
	const unsigned int address_len = (sizeof(edb::address_t) * CHAR_BIT) / 4;
	return address_len + (show_address_separator_ ? 1 : 0);
}

//------------------------------------------------------------------------------
// Name: instruction_size(quint8 *buffer, std::size_t size)
// Desc: 
//------------------------------------------------------------------------------
int QDisassemblyView::instruction_size(quint8 *buffer, std::size_t size) {
	edb::Instruction insn(buffer, size, 0, std::nothrow);
	return insn.size();
}

//------------------------------------------------------------------------------
// Name: 
// Desc: 
//------------------------------------------------------------------------------
edb::address_t QDisassemblyView::addressFromPoint(const QPoint &pos) const {
	const edb::address_t address = address_from_coord(pos.x(), pos.y()) + address_offset_;
	if(address >= region_.end) {
		return 0;
	}
	return address;
}

//------------------------------------------------------------------------------
// Name: 
// Desc: 
//------------------------------------------------------------------------------
int QDisassemblyView::get_instruction_size(edb::address_t address, bool &ok, quint8 *buf, int &size) const {
	int ret = 0;
	
	if(size < 0) {
		ok = false;
	} else {	
		ok = edb::v1::get_instruction_bytes(address, buf, size);
			
		if(ok) {
			ret = instruction_size(buf, size);
		}
	}
	return ret;
}

//------------------------------------------------------------------------------
// Name: get_instruction_size(edb::address_t address, bool &ok) const
// Desc: 
//------------------------------------------------------------------------------
int QDisassemblyView::get_instruction_size(edb::address_t address, bool &ok) const {
	quint8 buf[edb::Instruction::MAX_SIZE];
	
	// do the longest read we can while still not crossing region end
	int buf_size = sizeof(buf);			
	if(region_.end != 0 && address + buf_size > region_.end) {
		
		if(address <= region_.end) {
			buf_size = region_.end - address;
		} else {
			buf_size = 0;
		}
	}
	
	return get_instruction_size(address, ok, buf, buf_size);
}

//------------------------------------------------------------------------------
// Name: address_from_coord(int x, int y) const
// Desc: 
//------------------------------------------------------------------------------
edb::address_t QDisassemblyView::address_from_coord(int x, int y) const {
	Q_UNUSED(x);
	
	const int line = y / line_height();
	const edb::address_t offset = verticalScrollBar()->value();
	edb::address_t address = offset;

	// add up all the instructions sizes up to the line we want
	for(int i = 0; i < line; ++i) {
		bool ok;
				
		int size = get_instruction_size(address_offset_ + address, ok);
		if(ok) {
			address += (size != 0) ? size : 1;
		} else {
			address += 1;
		}
	}
	
	return address;
}

//------------------------------------------------------------------------------
// Name: mouseDoubleClickEvent(QMouseEvent *event)
// Desc: 
//------------------------------------------------------------------------------
void QDisassemblyView::mouseDoubleClickEvent(QMouseEvent *event) {
	if(event->button() == Qt::LeftButton) {
		if(event->x() < line1()) {
			const edb::address_t address = addressFromPoint(event->pos());
					
			if(region_.contains(address)) {
				emit breakPointToggled(address);
				repaint();
			}
		}
	}
}

//------------------------------------------------------------------------------
// Name: event(QEvent *event)
// Desc: 
//------------------------------------------------------------------------------
bool QDisassemblyView::event(QEvent *event) {

	if (event->type() == QEvent::ToolTip) {
		bool show = false;
	
		const QHelpEvent *const helpEvent = static_cast<QHelpEvent *>(event);

		if(helpEvent->x() >= line1() && helpEvent->x() < line2()) {
		
			const edb::address_t address = addressFromPoint(helpEvent->pos());
			
			quint8 buf[edb::Instruction::MAX_SIZE];
			
			// do the longest read we can while still not passing the region end
			int buf_size = qMin<edb::address_t>((region_.end - address), sizeof(buf));
			if(edb::v1::get_instruction_bytes(address, buf, buf_size)) {
				const edb::Instruction insn(buf, buf_size, address, std::nothrow);

				if((line1() + (static_cast<int>(insn.size()) * 3) * font_width_) > line2()) {
					const QString byteBuffer = format_instruction_bytes(insn);
					QToolTip::showText(helpEvent->globalPos(), byteBuffer);
					show = true;
				}
			}
		}
		
		if(!show) {
			QToolTip::showText(helpEvent->globalPos(), QString());
		}
	}
	
	return QAbstractScrollArea::event(event);
}

//------------------------------------------------------------------------------
// Name: mouseReleaseEvent(QMouseEvent *event)
// Desc: 
//------------------------------------------------------------------------------
void QDisassemblyView::mouseReleaseEvent(QMouseEvent *event) {

	Q_UNUSED(event);
	
	moving_line1_ = false;
	moving_line2_ = false;

	setCursor(Qt::ArrowCursor);
	repaint();
}

//------------------------------------------------------------------------------
// Name: updateSelectedAddress(QMouseEvent *event)
// Desc: 
//------------------------------------------------------------------------------
void QDisassemblyView::updateSelectedAddress(QMouseEvent *event) {
	bool ok;
	const edb::address_t address	= addressFromPoint(event->pos());
	const int size					= get_instruction_size(address, ok);

	if(ok) {
		selected_instruction_address_	= address;
		selected_instruction_size_		= size;
	} else {
		selected_instruction_address_	= 0;
		selected_instruction_size_		= 0;
	}
}

//------------------------------------------------------------------------------
// Name: near_line(int x, int linex)
// Desc: determins if a point is close to a column divider or not
//------------------------------------------------------------------------------
bool QDisassemblyView::near_line(int x, int linex) {
	return qAbs(x - linex) < 3;
}

//------------------------------------------------------------------------------
// Name: mousePressEvent(QMouseEvent *event)
// Desc: 
//------------------------------------------------------------------------------
void QDisassemblyView::mousePressEvent(QMouseEvent *event) {
	
	if(event->button() == Qt::LeftButton) {
		if(near_line(event->x(), line1())) {
			moving_line1_ = true;
		} else if(near_line(event->x(), line2())) {
			moving_line2_ = true;
		} else {
			updateSelectedAddress(event);
		}
	} else {
		updateSelectedAddress(event);
	}
}

//------------------------------------------------------------------------------
// Name: mouseMoveEvent(QMouseEvent *event)
// Desc: 
//------------------------------------------------------------------------------
void QDisassemblyView::mouseMoveEvent(QMouseEvent *event) {

	const int x_pos = event->x();

	if(moving_line1_) {
		if(x_pos >= auto_line1() && x_pos + font_width_ < line2()) {
			if(line2_ == 0) {
				line2_ = line2();
			}
			line1_ = x_pos;
		}
		repaint();
	} else if(moving_line2_) {
		if(x_pos > line1() + font_width_ && x_pos + 1 < width() - (verticalScrollBar()->width() + 3)) {
			line2_ = x_pos;
		}
		repaint();
	} else {
		if(near_line(x_pos, line1()) || near_line(x_pos, line2())) {
			setCursor(Qt::SplitHCursor);
		} else {
			setCursor(Qt::ArrowCursor);
		}
	}
}

//------------------------------------------------------------------------------
// Name: selectedAddress() const
// Desc: 
//------------------------------------------------------------------------------
edb::address_t QDisassemblyView::selectedAddress() const {
	return selected_instruction_address_;
}

//------------------------------------------------------------------------------
// Name: selectedSize() const
// Desc: 
//------------------------------------------------------------------------------
int QDisassemblyView::selectedSize() const {
	return selected_instruction_size_;
}

//------------------------------------------------------------------------------
// Name: region() const
// Desc: 
//------------------------------------------------------------------------------
MemRegion QDisassemblyView::region() const {
	return region_;
}
